///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of CRIME.                                       //
//                                                                   //
// CRIME is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// CRIME is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////
#include "common_gh.h"
#include <gsl/gsl_integration.h>
#include <gsl/gsl_errno.h>
#include "cosmo_mad.h"

//Cosmological parameters
double hhub=0.67; //h
double OmegaM=0.3;   //Omega_M
double OmegaL=0.7;   //Omega_L
double OmegaB=0.049; //Omega_b
double weos=-1;   //equation of state
double sig8=0.8;  //sigma_8 normalization
double n_scal=0.96;      //scalar spectral index
double fgrowth_0=0; //growth rate at z=0
double hubble_0=1; //Inverse horizon at z=0
double z_min=0; //Redshift range
double z_max=1.4;
double r_min;
double r_max;

//Power spectrum arrays
static int numk;
static double logkmax=3;
static double logkmin=-4;
static double idlogk=10;
static double *logkarr;
static double *pkarr;

//Arrays of z-dependent functions
#define DZ 0.001
#define NZ 5001

static double z_arr_z2r[NZ];
static double r_arr_z2r[NZ];
static double z_arr_r2z[NZ];
static double r_arr_r2z[NZ];
static double growth_d_arr[NZ];
static double growth_v_arr[NZ];

static double glob_idr;

static double z_of_r_provisional(double r)
{
  if(r<=0) return 0;
  else if(r>=r_arr_z2r[NZ-1]) return z_arr_z2r[NZ-1];
  else {
    int iz=0;
    while(r>=r_arr_z2r[iz])
      iz++;
    return z_arr_z2r[iz-1]+(z_arr_z2r[iz]-z_arr_z2r[iz-1])*
      (r-r_arr_z2r[iz-1])/(r_arr_z2r[iz]-r_arr_z2r[iz-1]);
  }
}

double r_of_z(double z)
{
  if(z<=0) return 0;
  else if(z>=z_arr_z2r[NZ-1]) return r_arr_z2r[NZ-1];
  else {
    int iz=(int)(z/DZ);
    double r=r_arr_z2r[iz]+(r_arr_z2r[iz+1]-r_arr_z2r[iz])*
      (z-z_arr_z2r[iz])/DZ;
    return r;
  }
}

double z_of_r(double r)
{
  if(r<=0) return 0;
  else if(r>=r_arr_r2z[NZ-1]) return z_arr_r2z[NZ-1];
  else {
    int ir=(int)(r*glob_idr);
    double z=z_arr_r2z[ir]+(z_arr_r2z[ir+1]-z_arr_r2z[ir])*
      (r-r_arr_r2z[ir])*glob_idr;
    return z;
  }
}

double dgrowth_of_r(double r)
{
  if(r<=0) return 1;
  else if(r>=r_arr_r2z[NZ-1]) return growth_d_arr[NZ-1];
  else {
    int ir=(int)(r*glob_idr);
    double gd=growth_d_arr[ir]+(growth_d_arr[ir+1]-growth_d_arr[ir])*
      (r-r_arr_r2z[ir])*glob_idr;
    return gd;
  }
}

double vgrowth_of_r(double r)
{
  if(r<=0) return 1;
  else if(r>=r_arr_r2z[NZ-1]) return growth_v_arr[NZ-1];
  else {
    int ir=(int)(r*glob_idr);
    double gv=growth_v_arr[ir]+(growth_v_arr[ir+1]-growth_v_arr[ir])*
      (r-r_arr_r2z[ir])*glob_idr;
    return gv;
  }
}

static void int_error_handle(int status,double result,
                             double error)
{
  //////
  // Error handler for gsl
  if(isnan(result)) {
    fprintf(stderr,"CRIME: NAN found \n");
    exit(1);
  }
  else{
    if(status==GSL_EROUND)
      fprintf(stderr,"CRIME: Roundoff error: %lE %lE \n",result,error);
    else if(status==GSL_EMAXITER)
      fprintf(stderr,"CRIME: Ran out of iterations: %lE %lE \n",result,error);
    else if(status==GSL_ESING)
      fprintf(stderr,"CRIME: Singularity found: %lE %lE \n",result,error);
    else if(status==GSL_EDIVERGE) {
      fprintf(stderr,"CRIME: Integral seems to diverge: %lE %lE \n",
	      result,error);
    }
    else if(status==GSL_ETOL) {
      fprintf(stderr,"CRIME: Can't reach tolerance: %lE %lE : %lE %%\n",
              result,error,100*error/result);
    }
    else if(status==GSL_EUNDRFLW)
      fprintf(stderr,"CRIME: Underflow: %lE %lE \n",result,error);
    else if(status==GSL_EDOM) {
      fprintf(stderr,"CRIME: Outside interpolation range!! %lE %lE\n",
	      result,error);
      exit(1);
    }
    else if(status) {
      fprintf(stderr,"CRIME: Unknown error code %d %lf %lf \n",
	      status,result,error);
      exit(1);
    }
  }
}

static double wind(double x,int setwf)
{
  //////
  // Window function:
  //  setwf=0 -> top hat
  //  setwf=1 -> gaussian
  //  setwf=2 -> sharp-k
  if(setwf==1) { //Gaussian
    return exp(-0.5*x*x);
  }
  else if(setwf==0) { //TopHat
    if(x<0.1) {
      return 1.-0.1*x*x+0.003571429*x*x*x*x-6.61376E-5*x*x*x*x*x*x
        +7.51563E-7*x*x*x*x*x*x*x*x;
    }
    else
      return 3*(sin(x)-x*cos(x))/(x*x*x);
  }
  else if(setwf==2) { //Sharp-k
    if(x<1) return 1;
    else return 0;
  }
  else
    return -1;
}

double pk_linear0(double lgk)
{
  //////
  // Linear power spectrum at redshift 0.
  // Extrapolated to ~k^ns for small k and
  // to k^{-3} for large k
  double pk;
  int ik=(int)((lgk-logkmin)*idlogk);
  
  if(ik<0)
    pk=pkarr[0]*pow(10,n_scal*(lgk-logkmin));
  else if(ik<numk)
    pk=pkarr[ik]+(lgk-logkarr[ik])*(pkarr[ik+1]-pkarr[ik])*idlogk;
  else
    pk=pkarr[numk-1]*pow(10,-3*(lgk-logkmax));

  return pk;
}

static double j_bessel_0(double x)
{
  //////
  // Bessel's j0 function
  if(x>0.1)
    return sin(x)/x;
  else
    return 1.-0.166667*x*x+0.00833333*x*x*x*x-
      0.000198413*x*x*x*x*x*x+2.75573E-6*x*x*x*x*x*x*x*x;
}

typedef struct { //Param struct for integrals
  double r;
  double R1;
  double R2;
  int wf1;
  int wf2;
} xiparam;

static double integxiL_O(double kk,void *params)
{
  //////
  // integrand for xi_L (for oscillatory integration)
  double dum;
  double x1,x2,xr;
  xiparam *par;
  double lgk=log10(kk);
  par=(xiparam *)params;

  x1=kk*(par->R1);
  x2=kk*(par->R2);
  xr=kk*(par->r);

  dum=TWOPIPIINV*pk_linear0(lgk)*kk*kk*
    wind(x1,par->wf1)*wind(x2,par->wf2)/xr;

  return dum;
}

static double integxiL_NO(double logk,void *params)
{
  //////
  // integrand for xi_L (including oscillatory term in j0)
  double dum;
  double x1,x2,xr;
  xiparam *par;
  par=(xiparam *)params;

  double kk=pow(10,logk);
  x1=kk*(par->R1);
  x2=kk*(par->R2);
  xr=kk*(par->r);

  dum=TWOPIPIINVLOGTEN*pk_linear0(logk)*kk*kk*kk*
    wind(x1,par->wf1)*wind(x2,par->wf2)*j_bessel_0(xr);

  return dum;
}

static double xi2p_L(double r,double R1,double R2,
		     char *wf1,char *wf2,double errfac)
{
  //////
  // Correlation function between the linear density contrast smoothed
  // with window function (wf1,R1) and with window function (wf2,R2)
  // at two points separated by a distance r:
  //              <delta_(R1,wf1)(x)*delta_(R2,wf2)(x+r)>
  gsl_function integrand;
  double relerrt=1E-4;
  double integral,errintegral;
  xiparam par;
  double lim=MIN(R1,R2);
  lim/=r;

  par.r=r;
  par.R1=R1;
  par.R2=R2;
  if(!strcmp(wf1,"Gauss"))
    par.wf1=1;
  else if(!strcmp(wf1,"TopHat"))
    par.wf1=0;
  else {
    fprintf(stderr,"CRIME: Unknown window function %s \n",wf1);
    exit(1);
  }
  if(!strcmp(wf2,"Gauss"))
    par.wf2=1;
  else if(!strcmp(wf2,"TopHat"))
    par.wf2=0;
  else {
    fprintf(stderr,"CRIME: Unknown window function %s \n",wf2);
    exit(1);
  }

  gsl_integration_workspace *w
    =gsl_integration_workspace_alloc(1000);
  integrand.params=&par;
  if(lim>=1) {
    integrand.function=&integxiL_NO;
    int stat=gsl_integration_qagil(&integrand,logkmax,0,relerrt,1000,w,
                                   &integral,&errintegral);
    int_error_handle(stat,integral,errintegral);
  }
  else {
    lim*=errfac;
    gsl_integration_workspace *cw
      =gsl_integration_workspace_alloc(1000);
    gsl_integration_qawo_table *wf
      =gsl_integration_qawo_table_alloc(r,0.1,GSL_INTEG_SINE,100);

    integrand.function=&integxiL_O;
    int stat=gsl_integration_qawf(&integrand,0,relerrt*lim,1000,
                           w,cw,wf,&integral,&errintegral);
    int_error_handle(stat,integral,errintegral);

    gsl_integration_qawo_table_free(wf);
    gsl_integration_workspace_free(cw);
  }
  gsl_integration_workspace_free(w);

  return integral;
}

static double sigL2(double R1,double R2,char *wf1,char *wf2)
{
  //////
  // Covariance between the linear density contrast smoothed with
  // window function (wf1,R1) and with window function (wf2,R2) at
  // the same point:  <delta_(R1,wf1)(x)*delta_(R2,wf2)(x)>
  return xi2p_L(0,R1,R2,wf1,wf2,1);
}

static void pk_linear_set(char *fname,double nns,double s8)
{
  //////
  // Reads linear power spectrum. CAMB format expected.
  int ii;
  double kk,ppk;
  FILE *fpk;

  n_scal=nns;
  sig8=s8;

  printf("Reading P_k from file: %s\n",fname);
  fpk=fopen(fname,"r");
  if(fpk==NULL) error_open_file(fname);

  numk=linecount(fpk);
  logkarr=(double *)my_malloc(numk*sizeof(double));
  pkarr=(double *)my_malloc(numk*sizeof(double));
  rewind(fpk);
  for(ii=0;ii<numk;ii++) {
    int stat=fscanf(fpk,"%lf %lf",&kk,&ppk);
    if(stat!=2) error_read_line(fname,ii+1);
    pkarr[ii]=ppk;
    logkarr[ii]=log10(kk); //log(k) in h Mpc^-1
  }
  logkmin=logkarr[0];
  logkmax=logkarr[numk-1];
  idlogk=(numk-1)/(logkarr[numk-1]-logkarr[0]);
  fclose(fpk);

  // normalize
  double norm_pk=sig8*sig8/sigL2(8,8,"TopHat","TopHat");
  printf("  Original sigma8=%lf\n",
	 sqrt(sigL2(8,8,"TopHat","TopHat")));
  for(ii=0;ii<numk;ii++)
    pkarr[ii]*=norm_pk;
}

void cosmo_set(void)
{
  //////
  // This initializes the cosmological model
  // at redshift z_s

  int ii;

  Csm_params *pars=csm_params_new();
  csm_unset_gsl_eh();
  csm_background_set(pars,OmegaM,OmegaL,OmegaB,weos,0,hhub,2.275);

  fgrowth_0=csm_f_growth(pars,1);
  hubble_0=csm_hubble(pars,1);

  z_min=NU_21/nu_max-1;
  z_max=NU_21/nu_min-1;
  r_min=csm_radial_comoving_distance(pars,1/(1+z_min));
  r_max=csm_radial_comoving_distance(pars,1/(1+z_max));

  l_box=2*r_max;
  pos_obs[0]=0.5*l_box;
  pos_obs[1]=0.5*l_box;
  pos_obs[2]=0.5*l_box;

  //Set z-dependent functions
  for(ii=0;ii<NZ;ii++) {
    double z=((double)ii)*DZ;
    double a=1/(1+z);
    double rz=csm_radial_comoving_distance(pars,a);
    z_arr_z2r[ii]=z;
    r_arr_z2r[ii]=rz;
  }
  if((z_arr_z2r[NZ-1]<=z_max)||(r_arr_z2r[NZ-1]<=r_max)) {
    fprintf(stderr,"CRIME: OMG!\n");
    exit(1);
  }
  
  double growth0=csm_growth_factor(pars,1);
  glob_idr=(NZ-1)/(r_arr_z2r[NZ-1]-r_arr_z2r[0]);
  for(ii=0;ii<NZ;ii++) {
    double r=ii/glob_idr;
    double z=z_of_r_provisional(r);
    double a=1/(1+z);
    z_arr_r2z[ii]=z;
    r_arr_r2z[ii]=r;

    double gz=csm_growth_factor(pars,a)/growth0;
    double fz=csm_f_growth(pars,a);
    double hhz=csm_hubble(pars,a);
    growth_d_arr[ii]=gz;
    //This is for the comoving velocity
    growth_v_arr[ii]=(gz*hhz*fz)/(fgrowth_0*hubble_0);
  }
  if((z_arr_r2z[NZ-1]<=z_max)||(r_arr_r2z[NZ-1]<=r_max)) {
    fprintf(stderr,"CRIME: OMG!\n");
    exit(1);
  }

#ifdef _DEBUG
  int nz=500;
  FILE *fil=fopen("test_cosmo.dat","w");
  for(ii=0;ii<nz;ii++) {
    double z1=5.*(ii+0.5)/nz;
    double r=csm_radial_comoving_distance(pars,1/(1+z1));
    double z2=z_of_r(r);
    double gz1=dgrowth_of_r(r);
    double gz2=csm_growth_factor(pars,1/(1+z1))/growth0;

    fprintf(fil,"%lE %lE %lE %d %lE %lE\n",z1,r,z2,nz,gz1,gz2);
  }
  fclose(fil);
#endif //_DEBUG
  csm_params_free(pars);

  pk_linear_set(fnamePk,n_scal,sig8);
}

void end_cosmo(void)
{
  //////
  // Frees up all memory and returns to original state
  free(pkarr);
  free(logkarr);
}
